using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Threading;
using System.Threading.Channels;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace ksp_osm
{
    public static class ParallelOsmReader
    {
                /// Paralelně vyčte všechny sekvence a spojí je do jedné
        public static IEnumerable<T> ParallelJoinSeqs<T>(IEnumerable<T>[] xss)
        {
            int ended = 0;
            var channel = System.Threading.Channels.Channel.CreateBounded<T>(new BoundedChannelOptions(10_000) { SingleWriter = false });
            var writer = channel.Writer;

            foreach (var xs in xss)
            {
                var t = new Thread(() => {
                    foreach (var x in xs)
                    {
                        channel.Writer.WriteAsync(x).AsTask().Wait();
                    }
                    Interlocked.Increment(ref ended);
                });
                t.Start();
            }

            while (ended < xss.Length)
            {
                if (channel.Reader.TryRead(out var x))
                {
                    yield return x;
                }
                else
                {
                    Thread.Sleep(20);
                }
            }
            while (channel.Reader.TryRead(out var x))
                yield return x;
            // foreach (var e in buffers.SelectMany(b => b))
            //     yield return e;
        }

        public static IEnumerable<XElement> LoadElements(string file, Func<string, Lazy<long>, bool> filter = null)
        {
            if (File.Exists(file + ".0"))
            {
                // Optimalizace - načítáme několik souborů najednou.
                // XML parser je totiž pěkně pomalý a nejde rozumně paralelizovat
                // 
                var filterLock = new object();
                Func<string, Lazy<long>, bool> lockedFilter = null;
                if (filter is object)
                    lockedFilter = (a, b) => {
                        lock (filterLock)
                            return filter(a, b);
                    };

                var n = 0;
                while (File.Exists(file + "." + n))
                    n++;
                Console.WriteLine($"Loading {n} files in parallel.");
                return ParallelJoinSeqs(Enumerable.Range(0, n).Select(i => OsmReader.LoadElements(OsmReader.OpenFile(file + "." + i), lockedFilter)).ToArray());
            }
            else
            {
                return OsmReader.LoadElements(OsmReader.OpenFile(file), filter);
            }
        }


        /// Utilita, která soubor rozdělí na několik, aby šel načítat parallelně
        public static void SplitDocument(string file, string outputFiles, int count)
        {
            Parallel.For(0, count, i => {
                long index = 0;
                var els =
                    ParallelOsmReader.LoadElements(file)
                    .Where((_) => index++ % count == i);
                var el = new XStreamingElement("osm", els);
                using (var ss = new GZipStream(File.Create(outputFiles + "." + i), CompressionMode.Compress))
                {
                    el.Save(ss);
                }
            });
        }
    }
}
